
var ttyPortId = -1;		// the opened tty port, we keep it so we can close it
var ttyReponse = '';	// the data received from the tty port
var ttyRxCount = 0;		// how many bytes received
var closeOnComplete = true;
var connectionCount = 0;


function stubStatusUpdate(statStr) {
	console.log('status:' + statStr);
}

var callWithStatus = stubStatusUpdate;

function closeTTY() {
	if(ttyPortId >= 0) {
		//chrome.serial.flush(ttyPortId, function(res) { });
		chrome.serial.disconnect(ttyPortId, function(res) { ttyPortId = -1;});
	}
	ttyReponse = '';
	ttyRxCount = 0;
	connectionCount = 0;
	// remove listeners, they'll be readded
	chrome.serial.onReceive.removeListener(gotSerialRxData);
	chrome.serial.onReceiveError.removeListener(gotSerialRxError);
}

function resetTTYRX() {
	ttyReponse = '';
	ttyRxCount = 0;
}

function ttyFailure(whence) {
	console.log('tty op failed at ' + whence);
	callWithStatus('TX failed at ' + whence);
	closeTTY();
	funcToReply(false, '');
}

function gotSerialRxData(info) {
	if (info.connectionId == ttyPortId) {
		var bb = new Uint8Array(info.data);
		for (i = 0; i < bb.byteLength; i++) {
			//console.log('rx[' + i + '] = ' + bb[i].toString() + ' ' + String.fromCharCode(bb[i]));
			ttyResponse += String.fromCharCode(bb[i]);
			ttyRxCount++;
		}
		//console.log('rx ' + ttyRxCount + ' ' + ttyResponse);
	}
	else {
		// happens for straggling data after we set port id 0
		//console.log('not for us?');
	}
}

function gotSerialRxError(info) {
	if (info.error != 'timeout') {
		console.log('tty rx err: ' + info.error);
		if ((cmdToSend.search('cyboot') < 0) && ttyPortId >= 0) {
			callWithStatus('Serial RX Error: ' + info.error);
		}
	}
}

function waitWhileRxData(whereNext, lastrx, curtimeout, initialtimeout) {

	// set a timer to wait 100ms and check rx count and if it changed,
	// set another timer, so as long as rxcount changes, we recurse into
	// this function (note that rx can go megabytes before this recursion
	// is an issue, but if so, set a longer timeout after a while
	//
	//console.log('while rx ' + curtimeout);
	
	setTimeout(function() {
		if (ttyRxCount != lastrx) {
			// more rx bytes after 100ms, so timeout more, and reset
			// quiet timeout to initial (restart end of rx timer)
			waitWhileRxData(whereNext, ttyRxCount, initialtimeout, initialtimeout);
		}
		else {
			// no data, start the next timeout 100ms less
			//if (curtimeout == initialtimeout)
			//	console.log('rx quiet to begins');
			if (curtimeout < 100)
				whereNext(true, ttyResponse);
			else
				waitWhileRxData(whereNext, lastrx, curtimeout - 100, initialtimeout);
		}
	}, 50);
}

function waitForRxData(whereNext, expectData, timeoutfirst, timeoutlast) {
	
	// set a timer to wait for at least timeout milliseconds
	// for Rx data count to change, and then continue
	// the timeout is divided into 100ms segments so the 
	// next step is started quicker
	//
	setTimeout(function() {
		if (ttyRxCount == 0) {
			// no rx bytes after 100ms, so timeout again, but for shorter time
			if (timeoutfirst < 100) {
				if (expectData)
					console.log('RX timeout');
				whereNext(expectData ? false : true, '');
			}
			else
				waitForRxData(whereNext, expectData, timeoutfirst - 100, timeoutlast);
		}
		else
			waitWhileRxData(whereNext, ttyRxCount, timeoutlast, timeoutlast);
	}, 100);

	// note that this return happens instantly (the timeout function is run
	// in the javacript thread when the timer expires).  this is what we want
	// as it allows js to fetch the serial data
}

function sendTTYraw(bytes, checkRx, andDoThisNext)
{
	txComplete = false;

	chrome.serial.send(ttyPortId, bytes, function(sendInfo) {
		//console.log('sent ' + sendInfo.bytesSent);
		chrome.serial.flush(ttyPortId, function(res) {
			//console.log('flush='+res);
			// if didn't send whole string, error out
			if (sendInfo.bytesSent != bytes.byteLength) {
				ttyFailure('sending raw ' + sendInfo.bytesSent + ' != ' + bytes.byteLength);
				andDoThisNext(false, '');
			} else {
				if (checkRx && (ttyRxCount > 0))
					// check rx reply, but only with a short timer
					waitForRxData(andDoThisNext, false, 0, 100);
				else
					andDoThisNext(true, '');
			}
		});
	});
}

function sendTTY(str, whereNext)
{
	var bytes = new ArrayBuffer(str.length);
	var bb = new Uint8Array(bytes);
	
	for (var i = 0; i < str.length; i++)
	{
	    bb[i] = str.charCodeAt(i);
	}
	ttyResponse = '';
	ttyRxCount = 0;

	chrome.serial.send(ttyPortId, bytes, function(sendInfo) {
		//console.log('sent ' + sendInfo.bytesSent);
		chrome.serial.flush(ttyPortId, function(res) {
			//console.log('flush='+res);
			// if didn't send whole string, error out
			if (sendInfo.bytesSent != str.length) {
				ttyFailure('sending ' + sendInfo.bytesSent + ' != ' + str.length);
				whereNext(false, '');
			} else {
				// after sending bytes, timeout waiting for rx.  expect the first rx within 
				// a second but wait 8, and we wait 1/4 second after rx quiet to make sure its all there
				//
				var firstWait = 10000; // up to 10 seconds for first byte of reply
				var dwellWait = 350;   // 1/4 second dwell after last byte rx
				
				if (str.search('flash') >= 0) {
					// if flashing, full response can be a while, so don't even start
					// looking for first byte for a bit
					//
					dwellWait = 1500;
					setTimeout(function() {
							waitForRxData(whereNext, true, firstWait, dwellWait);
						},
						1500);
				} else if(str.search('cyboot') >= 0) {
					// if going to usb bootloader mode, that's a re-enum
					// so don't look for rx results, just go
					whereNext(true, '');
				} else {
					waitForRxData(whereNext, true, firstWait, dwellWait);
				}
			}
		});
	});
}

function logoutMode(rxdata) {
	closeTTY();
		
	// set hid mode back to normal (hide cdc)
	//
	hidSetMode(USB_MODE_NORMAL);
}

function getCommandResponse(laststat, rxdata) {
	// copy and trim response, so it can be stomped on by sendTTY
	response = rxdata.trim();
	//console.log('resp for ' + cmdToSend + ' is ' + response);
	if (closeOnComplete) {
		// logout, and close TTY if request set islast (command/req)
		// and do this *before* calling back because that might
		// chain the next req and change the global closeOnComplete
		console.log('closing tty on complete\n');
		sendTTY('exit\r', logoutMode);
	}
	funcToReply(laststat, response);
}

function doSendCommand() {
	console.log('M-Send command: ' + cmdToSend);
	sendTTY(cmdToSend, getCommandResponse);
}

function checkLoggedIn() {
	console.log('M-Logged In?:' + ttyResponse);
	
	// if the response string is "#" then in login mode, advance
	var pos = ttyResponse.search('#');
	
	if (pos >= 0) {
		doSendCommand();
	}
	else {
		ttyFailure('checked logged in');
	}
}

function passwordMode() {	
	console.log('M-Password:' + ttyResponse);
	
	// if the response string is "#" then in login mode, advance
	var pos = ttyResponse.search('Password:');
	
	if (pos >= 0) {
		// after logging in, we expect to be in login mode, so 
		// we could technically go back to initial prompt mode
		// and springboard from there to doCommand but that could
		// loop, so I use a separate prompt mode here
		//
		sendTTY('ben1jar\r', checkLoggedIn);
	}
	else {
		ttyFailure('password');
	}
}

function loginMode() {	
	console.log('M-Login:' + ttyResponse);
	
	var pos = ttyResponse.search('login:');
	
	if (pos >= 0) {
		sendTTY('root\r', passwordMode);
	}
	else {
		ttyFailure('login');
	}
}

function toRubyMode() {	
	console.log('M-ToRuby:' + ttyResponse);
	
	var pos = ttyResponse.search('ruby>');
	
	if (pos >= 0) {
		doSendCommand();
	}
	else {
		ttyFailure('login');
	}
}

function getInitialPrompt() {	
	// see what prompt we want
	var rubycon = (cmdLoginLevel == 'ruby') ? true : false;
	
	// remove white space (CR-LF, etc.) from both ends
	var respStr = ttyResponse.trim();
	
	// if the response string is "#" then in login mode, advance
	var pos = respStr.search('#');

	console.log('M-initial Prompt:' + ttyResponse);
	
	if (pos >= 0 && pos < 2) {
		console.log('already logged in mode');
		if (rubycon) {
			// want to be in rubycon, so exit
			sendTTY('exit\r', toRubyMode);
		} else {
			doSendCommand();
		}
		return;
	}
	// if the response string is "ruby>" then goto login mode
	pos = respStr.search('ruby>');
	if (pos >= 0) {
		if (rubycon) {
			// got to rubycon, all we need so do it now
			console.log('X-ruby mode, do cmd');
			doSendCommand();
			return;
		} else {
			console.log('X-ruby mode, login');
			sendTTY('quit\r', loginMode);
			return;
		}
	}			
	pos = respStr.search('ruby login:');
	if (pos >= 0) {
		console.log('X-ruby mode, login prompt');
		loginMode();
		return;
	}
	pos = respStr.search('Password:');
	if (pos >= 0) {
		console.log('X-ruby mode, password prompt');
		passwordMode();
		return;
	}
	ttyFailure('get initial prompt');
}

function onPortConnectedForCommand(coninfo) {
	console.log('tty connected', coninfo.connectionId);
	callWithStatus('Connected');
	ttyPortId = coninfo.connectionId;
	ttyReponse = '';
	ttyRxCount = 0;

	if(connectionCount == 0) {
		chrome.serial.onReceive.addListener(gotSerialRxData);
		chrome.serial.onReceiveError.addListener(gotSerialRxError);
	}
	connectionCount++;

	// first determine what state the device is in by sending
	// a single CR. We expect it to be in "ruby>" state but
	// it could possibly be in "#" state or unresponsive.
	//
	sendTTY('\r', getInitialPrompt)
}

function runTTYcommandAfterLogin(command, level, statusUpdater, islast, resultfunc) {
	// have to switch to globals now, the .connect call should really take
	// a parameter to it.
	if (statusUpdater != null)
		callWithStatus = statusUpdater;
	else
		callWithStatus = stubStatusUpdate;
	closeOnComplete = islast;
	cmdLoginLevel	= level;
	cmdToSend       = command;
	funcToReply     = resultfunc;
	
	//console.log('runcmd, connections=' + connectionCount);
	console.log('runcmd ' + command);

	if (ttyPortId >= 0) {
		// serial port already open, reuse it
		console.log('Already open, reuse port');
		sendTTY('\r', getInitialPrompt)
	}
	else {
		chrome.serial.getDevices(function(ports) {
			// enumerate ports
			for (i = 0; i < ports.length; i++) {
				//console.log('cport: ' + ports[i].path + '[' + ports[i].vendorId + ':' + ports[i].productId + ']');
				if (ports[i].vendorId == USB_VENDOR_ID && ports[i].productId == USB_PRODUCT_ID) {
					var seropts = { "bitrate": 230400, "receiveTimeout": 1000 };
					
					chrome.serial.connect(ports[i].path, seropts, onPortConnectedForCommand);
					if (chrome.runtime.lastError) {
						console.log('tty connect failed:' + chrome.runtime.lastError.message);
						laststat= false;
					}
					return;
				}
			}
		});
	}
}

function runTTYcommandInTTYmode(attempt, command, level, statusUpdater, islast, resultfunc) {

	if (attempt++ > 3) {
		console.log('no tty port');
		if (statusUpdater)
			statusUpdater('-No TTY port');
		return;
	}
	var gotPortTimer = null;
	var gotHidTimer = null;
	
	console.log('look for serial port');
	chrome.serial.getDevices(function(ports) {
		// enumerate ports
		for (i = 0; i < ports.length; i++) {
			//console.log('tport: ' + ports[i].path + '[' + ports[i].vendorId + ':' + ports[i].productId + ']');
			if (ports[i].vendorId == USB_VENDOR_ID && ports[i].productId == USB_PRODUCT_ID) {
				// found our tty for our vid/pid
				if (gotHidTimer != null)
					clearTimeout(gotHidTimer);
				if (gotPortTimer != null)
					clearTimeout(gotPortTimer);
				runTTYcommandAfterLogin(command, level, statusUpdater, islast, resultfunc);
				return;
			}
		}
	});
	// start a long timeout waiting for serial device enumeration. if found, it
	// will cancel this timeout, so it won't really be long.  we can't call
	// serial.getDevices recursively it seems
	//
	gotPortTimer = setTimeout(function() {
			console.log('switch to ttymode');
				
			// make sure we get to cdc mode (doesn't do anything if already in mode)
			//
			closeTTY();
			hidSetMode(USB_MODE_PC);
			
			// some systems take a lot longer than others to be ready after CDC enumeration
			//
			var renum_delay = 4000;

			chrome.runtime.getPlatformInfo(function(info) {
			    // Display host OS in the console
			    //console.log(info.os);
				if (info.os == 'linux') 
					renum_delay = 20000;
	
				gotHidTimer = setTimeout(function() {
						// recurse into here, with attempt incremented
						console.log('retry tty');
						runTTYcommandInTTYmode(attempt, command, level, statusUpdater, islast, resultfunc);
					},
					// on linux desktop it takes a good 20 seconds after re-enum before
					// serial port works right
					//
					renum_delay
				);
			});
		},
		4000
	);
}

function getFileContents(filename, statusUpdater, islast, resultfunc) {
	// run "cat filename" in a logged in context
	// which sets the filecontent var to the result
	//
	runTTYcommandInTTYmode(1, 'cat ' + filename + '\r', 'login', statusUpdater, islast, resultfunc);
}

function runCommand(command, loginLevel, statusUpdater, islast, resultfunc) {
	runTTYcommandInTTYmode(1, command + '\r', loginLevel, statusUpdater, islast, resultfunc);
}

